# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

from munkres import Munkres


def distance_function_match(l1, l2, thresh, dist_fn, norm_funcs=[]):
    """Returns pairs of matching indices from l1 and l2."""
    common = []
    # We will keep track of the global index in the source list as we
    # will successively reduce their sizes.
    l1 = list(enumerate(l1))
    l2 = list(enumerate(l2))

    # Use the distance function and threshold on hints given by normalization.
    # See _match_by_norm_func for implementation details.
    # Also wrap the list element function function to ignore the global list
    # index computed above.
    for norm_fn in norm_funcs:
        new_common, l1, l2 = _match_by_norm_func(
                l1, l2,
                lambda a: norm_fn(a[1]),
                lambda a1, a2: dist_fn(a1[1], a2[1]),
                thresh)
        # Keep only the global list index in the end result.
        common.extend((c1[0], c2[0]) for c1, c2 in new_common)

    # Take any remaining umatched entries and try to match them using the
    # Munkres algorithm.
    dist_matrix = [[dist_fn(e1, e2) for i2, e2 in l2] for i1, e1 in l1]

    # Call Munkres on connected components on the remaining bipartite graph.
    # An edge links an element from l1 with an element from l2 only if
    # the distance between the elements is less (or equal) than the theshold.
    components = BipartiteConnectedComponents()
    for l1_i in range(len(l1)):
        for l2_i in range(len(l2)):
            if dist_matrix[l1_i][l2_i] > thresh:
                continue
            components.add_edge(l1_i, l2_i)

    for l1_indices, l2_indices in components.get_connected_components():
        # Build a partial distance matrix for each connected component.
        part_l1 = [l1[i] for i in l1_indices]
        part_l2 = [l2[i] for i in l2_indices]

        part_dist_matrix = [[dist_matrix[l1_i][l2_i] for l2_i in l2_indices]
                            for l1_i in l1_indices]
        part_cmn = _match_munkres(part_l1, part_l2, part_dist_matrix, thresh)

        common.extend((c1[0], c2[0]) for c1, c2 in part_cmn)

    return common


def _match_by_norm_func(l1, l2, norm_fn, dist_fn, thresh):
    """Matches elements in l1 and l2 using normalization functions.

    Splits the elements in each list into buckets given by the normalization
    function. If the same normalization value points to a bucket from the
    first list and a bucket from the second list, both with a single element
    we consider the elements in the list as matching if the distance between
    them is less (or equal) than the threshold.

    e.g. l1 = ['X1', 'Y1', 'Y2', 'Z5'], l2 = ['X1', 'Y3', 'Z1']
         norm_fn = lambda x: x[0]
         dist_fn = lambda e1, e2: 0 if e1 == e2 else 1
         thresh = 0

    The buckets will then be:
        l1_bucket = {'X': ['X1'], 'Y': ['Y1', 'Y2'], 'Z': ['Z5']}
        l2_bucket = {'X': ['X1'], 'Y': ['Y3'], 'Z': ['Z1']}

    For each normalized value:
        'X' -> consider 'X1' equal with 'X1' since the distance is equal with
               the thershold
        'Y' -> skip the lists since we have multiple possible matches
        'Z' -> consider 'Z1' and 'Z5' as different since the distance is
               greater than the threshold.
    Return:
        [('X1', 'X2')]
    """
    common = []

    l1_only_idx = set(range(len(l1)))
    l2_only_idx = set(range(len(l2)))

    buckets_l1 = _group_by_fn(enumerate(l1), lambda x: norm_fn(x[1]))
    buckets_l2 = _group_by_fn(enumerate(l2), lambda x: norm_fn(x[1]))

    for normed, l1_elements in buckets_l1.items():
        l2_elements = buckets_l2.get(normed, [])
        if not l1_elements or not l2_elements:
            continue
        _, (_, e1_first) = l1_elements[0]
        _, (_, e2_first) = l2_elements[0]
        match_is_ambiguous = not (
            len(l1_elements) == len(l2_elements) and (
                all(e2 == e2_first for (_, (_, e2)) in l2_elements) or
                all(e1 == e1_first for (_, (_, e1)) in l1_elements)
            )
        )
        if match_is_ambiguous:
            continue
        for (e1_idx, e1), (e2_idx, e2) in zip(l1_elements, l2_elements):
            if dist_fn(e1, e2) > thresh:
                continue
            l1_only_idx.remove(e1_idx)
            l2_only_idx.remove(e2_idx)
            common.append((e1, e2))

    l1_only = [l1[i] for i in l1_only_idx]
    l2_only = [l2[i] for i in l2_only_idx]

    return common, l1_only, l2_only


def _match_munkres(l1, l2, dist_matrix, thresh):
    """Matches two lists using the Munkres algorithm.

    Returns pairs of matching indices from the two lists by minimizing the sum
    of the distance between the linked elements and taking only the elements
    which have the distance between them less (or equal) than the threshold.
    """
    equal_dist_matches = set()
    m = Munkres()
    indices = m.compute(dist_matrix)

    for l1_idx, l2_idx in indices:
        dst = dist_matrix[l1_idx][l2_idx]
        if dst > thresh:
            continue
        for eq_l2_idx, eq_val in enumerate(dist_matrix[l1_idx]):
            if abs(dst - eq_val) < 1e-9:
                equal_dist_matches.add((l1_idx, eq_l2_idx))
        for eq_l1_idx, eq_row in enumerate(dist_matrix):
            if abs(dst - eq_row[l2_idx]) < 1e-9:
                equal_dist_matches.add((eq_l1_idx, l2_idx))

    return [(l1[l1_idx], l2[l2_idx]) for l1_idx, l2_idx in equal_dist_matches]


class BipartiteConnectedComponents(object):
    """Union-Find implementation for getting connected components."""

    def __init__(self):
        self.parents = {}

    def add_edge(self, p1_node, p2_node):
        node_1 = (p1_node, None)
        node_2 = (None, p2_node)
        self._union(node_1, node_2)

    def get_connected_components(self):
        components_by_root = _group_by_fn(self.parents.keys(),
                                          self._find)
        for root in components_by_root:
            components_by_root[root].append(root)

        for component in components_by_root.values():
            p1_nodes = []
            p2_nodes = []
            for p1_node, p2_node in component:
                if p1_node is None:
                    p2_nodes.append(p2_node)
                if p2_node is None:
                    p1_nodes.append(p1_node)
            yield (p1_nodes, p2_nodes)

    def _union(self, node_1, node_2):
        parent_1 = self._find(node_1)
        parent_2 = self._find(node_2)
        if parent_1 != parent_2:
            self.parents[parent_1] = parent_2

    def _find(self, node):
        root = node
        while root in self.parents:
            root = self.parents[root]
        while node in self.parents:
            prev_node = node
            node = self.parents[node]
            self.parents[prev_node] = root
        return root


def _group_by_fn(iterable, fn):
    buckets = {}
    for elem in iterable:
        key = fn(elem)
        if key not in buckets:
            buckets[key] = []
        buckets[key].append(elem)
    return buckets



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Test dictdiffer wrappers."""

from __future__ import absolute_import, print_function

import pytest

from json_merger.config import DictMergerOps
from json_merger.conflict import Conflict, ConflictType
from dictdiffer.conflict import Conflict as Dictdiffer_Conflict
from json_merger.dict_merger import SkipListsMerger
from json_merger.errors import MergeError
from json_merger.nothing import NOTHING


def test_simple_behavior():
    r = {}
    h = {'foo': 'bar'}
    u = {'bar': 'baz'}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    assert m.merged_root == {'foo': 'bar', 'bar': 'baz'}


def test_base_values():
    m = SkipListsMerger(1, 2, 1, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == 2

    m = SkipListsMerger(1, 1, 2, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == 2


def test_base_values_exceptions():
    m = SkipListsMerger(1, 3, 2, DictMergerOps.FALLBACK_KEEP_HEAD)
    with pytest.raises(MergeError):
        m.merge()
    assert m.conflicts[0] == Conflict(ConflictType.SET_FIELD, (), 2)
    assert m.merged_root == 3

    m = SkipListsMerger(1, 3, 2, DictMergerOps.FALLBACK_KEEP_UPDATE)
    with pytest.raises(MergeError):
        m.merge()
    assert m.conflicts[0] == Conflict(ConflictType.SET_FIELD, (), 3)
    assert m.merged_root == 2


def test_merge_with_nothing():
    m = SkipListsMerger(1, {'some': 'other object'}, NOTHING,
                        DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == {'some': 'other object'}

    m = SkipListsMerger(1, NOTHING, {'some': 'other object'},
                        DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == {'some': 'other object'}

    m = SkipListsMerger(NOTHING, {'some': 'other object'}, NOTHING,
                        DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == {'some': 'other object'}


def test_simple_conflicts_keep_head():
    r = {}
    h = {'foo': 'bar'}
    u = {'foo': 'baz'}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == {'foo': 'bar'}
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 1
    assert Conflict(ConflictType.SET_FIELD, ('foo', ), 'baz') in m.conflicts


def test_simple_conflicts_keep_update():
    r = {}
    h = {'foo': 'bar'}
    u = {'foo': 'baz'}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_UPDATE)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == {'foo': 'baz'}
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 1
    assert Conflict(ConflictType.SET_FIELD, ('foo', ), 'bar') in m.conflicts


def test_simple_remove_conflict():
    r = {'foo1': 'bar', 'foo2': 'bar'}
    h = {'foo1': 'baz', 'foo2': 'baz'}
    u = {}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == {'foo1': 'baz', 'foo2': 'baz'}
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 2
    assert Conflict(ConflictType.REMOVE_FIELD, ('foo1', ), None) in m.conflicts
    assert Conflict(ConflictType.REMOVE_FIELD, ('foo2', ), None) in m.conflicts


# custom ops tests

def test_conflict_with_custom_ops_update():
    r = {'r': {'foo': 'baa', 'spam': 'eg'},
         'p': {'foo': 'baa', 'spam': 'eg'}}
    h = {'r': {'foo': 'bab', 'spam': 'egg'},
         'p': {'foo': 'bab', 'spam': 'egg'}}
    u = {'r': {'foo': 'bac', 'spam': 'eggs'},
         'p': {'foo': 'bac', 'spam': 'eggs'}}

    # set different strategies compare to the default one
    custom_ops = {
        'r': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p': DictMergerOps.FALLBACK_KEEP_UPDATE
    }
    expected = {
        'r': {'foo': 'bac', 'spam': 'eggs'},
        'p': {'foo': 'bac', 'spam': 'eggs'}
    }

    m = SkipListsMerger(
        r, h, u,
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == expected
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 4
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bab'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bab'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bab'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bab'
    ) in m.conflicts


def test_conflict_with_custom_ops_update_and_head():
    r = {'r': {'foo': 'baa', 'spam': 'eg'},
         'p': {'foo': 'baa', 'spam': 'eg'}}
    h = {'r': {'foo': 'bab', 'spam': 'egg'},
         'p': {'foo': 'bab', 'spam': 'egg'}}
    u = {'r': {'foo': 'bac', 'spam': 'eggs'},
         'p': {'foo': 'bac', 'spam': 'eggs'}}

    # set custom mixed strategies
    custom_ops = {
        'r': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p': DictMergerOps.FALLBACK_KEEP_HEAD,
    }
    expected = {
        'r': {'foo': 'bac', 'spam': 'eggs'},
        'p': {'foo': 'bab', 'spam': 'egg'}
    }

    m = SkipListsMerger(
        r, h, u,
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == expected
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 4
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'spam'), 'eggs'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'spam'), 'egg'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'foo'), 'bab'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bac'
    ) in m.conflicts


def test_conflict_with_custom_ops_update_and_head_mixed():
    r = {'r': {'foo': 'baa', 'spam': 'eg'},
         'p': {'foo': 'baa', 'spam': 'eg'}}
    h = {'r': {'foo': 'bab', 'spam': 'egg'},
         'p': {'foo': 'bab', 'spam': 'egg'}}
    u = {'r': {'foo': 'bac', 'spam': 'eggs'},
         'p': {'foo': 'bac', 'spam': 'eggs'}}

    # set custom mixed strategies with same root
    custom_ops = {
        'r': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p.foo': DictMergerOps.FALLBACK_KEEP_HEAD,
    }
    expected = {
        'r': {'foo': 'bac', 'spam': 'eggs'},
        'p': {'foo': 'bab', 'spam': 'eggs'}
    }

    m = SkipListsMerger(
        r, h, u,
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == expected

    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 4
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'spam'), 'egg'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'spam'), 'egg'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'foo'), 'bab'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bac'
    ) in m.conflicts


def test_conflict_with_custom_ops_update_and_head_with_nested_rules():
    r = {'r': {'foo': 'baa', 'spam': 'eg'},
         'p': {'foo': 'baa', 'spam': 'eg'}}
    h = {'r': {'foo': 'bab', 'spam': 'egg'},
         'p': {'foo': 'bab', 'spam': 'egg'}}
    u = {'r': {'foo': 'bac', 'spam': 'eggs'},
         'p': {'foo': 'bac', 'spam': 'eggs'}}

    custom_ops = {
        'r': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p': DictMergerOps.FALLBACK_KEEP_HEAD,
        'r.foo': DictMergerOps.FALLBACK_KEEP_HEAD,
        'p.spam': DictMergerOps.FALLBACK_KEEP_UPDATE
    }
    expected = {
        'r': {'foo': 'bab', 'spam': 'eggs'},
        'p': {'foo': 'bab', 'spam': 'eggs'}
    }

    m = SkipListsMerger(
        r, h, u,
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == expected

    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 4
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'spam'), 'egg'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'spam'), 'egg'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'foo'), 'bac'
    ) in m.conflicts
    assert Conflict(
        ConflictType.SET_FIELD, ('p', 'foo'), 'bac'
    ) in m.conflicts


def test_custom_fallback():
    r = {'r': {'foo': 'baa'}}
    h = {'r': {'foo': 'bab'}}
    u = {'r': {'foo': 'bac'}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == {'r': {'foo': 'bab'}}
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 1
    assert Conflict(
        ConflictType.SET_FIELD, ('r', 'foo'), 'bac'
    ) in m.conflicts


# Tests for all the list cases.

def test_head_only_list_add_no_skipped_lists():
    r = {'r': {'x': 1}}
    h = {'r': {'x': 1, 'l': [1, 2, 3]}}
    u = {'r': {'x': 2}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    assert m.merged_root == {'r': {'x': 2, 'l': [1, 2, 3]}}
    assert len(m.skipped_lists) == 0


def test_head_and_update_list_add_skipped_lists():
    r = {'r': {'x': 1}}
    h = {'r': {'x': 1, 'l': [1, 2, 3]}}
    u = {'r': {'x': 2, 'l': [1, 2, 3]}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    assert m.merged_root == {'r': {'x': 2}}
    # Knowing the list backup method check that m.merge is idempotent.
    assert r == r
    assert h == h
    assert u == u
    assert m.skipped_lists == set([('r', 'l')])


def test_update_deletes_root_list_no_conflict():
    r = {'r': {'x': 1, 'l': [1, 2, 3]}}
    h = {'r': {'x': 1, 'l': [1, 2, 3]}}
    u = {'r': {'x': 2}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    assert m.merged_root == {'r': {'x': 2}}
    assert len(m.skipped_lists) == 0


def test_one_list_delete_touched_in_head_raises_conflict():
    r = {'r': {'x': 1, 'l': [1, 2, 3]}}
    h = {'r': {'x': 1, 'l': [4, 3, 2, 1]}}
    u = {'r': {'x': 2}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == {'r': {'x': 2, 'l': [4, 3, 2, 1]}}
    assert len(m.skipped_lists) == 0
    assert m.conflicts == excinfo.value.content
    assert len(m.conflicts) == 1
    assert Conflict(ConflictType.REMOVE_FIELD, ('r', 'l'), None) in m.conflicts


def test_two_list_edit_skipped_lists():
    r = {'r': {'x': 1, 'l1': [1, 2, 3], 'l2': [1]}}
    h = {'r': {'x': 1, 'l1': [4, 3, 2, 1], 'l2': [2]}}
    u = {'r': {'x': 2, 'l1': [1, 2, 3], 'l2': [2]}}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    # The lists are kept as seen in root
    assert m.merged_root == {'r': {'x': 2, 'l1': [1, 2, 3], 'l2': [1]}}
    assert m.skipped_lists == set([('r', 'l1'), ('r', 'l2')])
    # Knowing the list backup method check that m.merge is idempotent.
    assert r == r
    assert h == h
    assert u == u


def test_data_lists_nested():
    r = {'r': [[1, 2, 3], [4, 5, 6]]}
    h = {'r': [[1, 2, 3], [4, 5, 6]]}
    u = {'r': [[3, 3, 3], [4, 5, 6]]}

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD,
                        data_lists=['r'])
    m.merge()
    assert m.merged_root == u
    assert not m.skipped_lists


def test_data_lists_bare_lists():
    r = [[1, 2, 3], [4, 5, 6]]
    h = [[1, 2, 3], [4, 5, 6]]
    u = [[3, 3, 3], [4, 5, 6]]

    m = SkipListsMerger(r, h, u, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()
    assert m.merged_root == u
    assert not m.skipped_lists


def test_get_custom_strategy():
    patches = [
        Dictdiffer_Conflict(
            ('change', 'p.foo', ('baa', 'bab')),
            ('change', 'p.foo', ('baa', 'bac'))
        ),
        Dictdiffer_Conflict(
            ('change', 'p.spam', ('eg', 'egg')),
            ('change', 'p.spam', ('eg', 'eggs'))
        ),
        Dictdiffer_Conflict(
            ('change', 'r.foo', ('baa', 'bab')),
            ('change', 'r.foo', ('baa', 'bac'))
        ),
        Dictdiffer_Conflict(
            ('change', 'r.spam', ('eg', 'egg')),
            ('change', 'r.spam', ('eg', 'eggs'))
        )
    ]

    custom_ops = {
        'r': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'p': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategies = ['s', 's', 's', 's']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_custom_strategy_mixed_patches():
    patches = [
        Dictdiffer_Conflict(
            ('change', 'b', ('1', '2')),
            ('remove', '', [('b', '1')])
        ),
        Dictdiffer_Conflict(
            ('add', '', [('a', '1')]),
            ('add', '', [('a', 'd')])
        )
    ]

    custom_ops = {
        'b': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategies = ['s', 's']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_custom_strategy_mixed_patches_with_nested_changes():
    patch = Dictdiffer_Conflict(
        ('change', 'a.b', ('1', '2')),
        ('change', 'a.b', ('1', '3'))
    )

    custom_ops = {
        'a.b': DictMergerOps.FALLBACK_KEEP_UPDATE,
    }

    expected_strategy = 's'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_change_remove_patch():
    patch = Dictdiffer_Conflict(
        ('change', 'a.b', ('1', '2')),
        ('remove', '', [('a', {'b': '1'})])
    )

    custom_ops = {
        'a.b': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategy = 's'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_change_remove_add_patches_multiple_changes():
    patches = [
        Dictdiffer_Conflict(
            ('change', 'a.c', ('2', '4')),
            ('remove', '', [('a', {'c': '2', 'b': '1'})])
        ),
        Dictdiffer_Conflict(
            ('change', 'a.b', ('1', '3')),
            ('remove', '', [('a', {'c': '2', 'b': '1'})])
        ),
        Dictdiffer_Conflict(
            ('add', 'a', [('d', '5')]),
            ('remove', '', [('a', {'c': '2', 'b': '1'})])
        ),
        Dictdiffer_Conflict(
            ('add', 'a', [('z', '7')]),
            ('remove', '', [('a', {'c': '2', 'b': '1'})])
        )
    ]

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategies = ['s', 's', 's', 's']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_custom_strategy_add_patch_very_nested():
    patch = Dictdiffer_Conflict(
        ('add', 'a.b.c', [('e', '2')]),
        ('add', 'a.b.c', [('e', '3')]))

    custom_ops = {
        'b': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategy = 's'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_add_change_patches_very_nested():
    patches = [
        Dictdiffer_Conflict(
            ('change', 'a.b.c.d', ('1', '4')),
            ('change', 'a.b.c.d', ('1', '5'))
        ),
        Dictdiffer_Conflict(
            ('add', 'a.b.c', [('e', '2')]),
            ('add', 'a.b.c', [('e', '3')])
        )
    ]

    custom_ops = {
        'b': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategies = ['s', 's']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_custom_strategy_add_patch_in_the_root():
    patch = Dictdiffer_Conflict(
        ('add', '', [('a', '1')]),
        ('add', '', [('a', '2')])
    )

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    expected_strategy = 's'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_mixed_patch_9():
    patch = Dictdiffer_Conflict(
        ('add', 'a', [('b', '1')]),
        ('add', 'a', [('b', '2')])
    )

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a.b': DictMergerOps.FALLBACK_KEEP_HEAD
    }

    expected_strategy = 'f'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_mixed_patch_with_list():
    patch = Dictdiffer_Conflict(
        ('add', ['a', 'c', 0], [('b', '1')]),
        ('add', ['a', 'c', 0], [('b', '2')])
    )

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a.c.b': DictMergerOps.FALLBACK_KEEP_HEAD
    }

    expected_strategy = 'f'

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategy = m._get_custom_strategy(patch)

    assert expected_strategy == strategy


def test_get_custom_strategy_mixed_patches_multiple_changes_with_list():
    patches = [
        Dictdiffer_Conflict(
            ('add', ['a', 'c', 0], [('b', '1')]),
            ('add', ['a', 'c', 0], [('b', '2')])
        ),
        Dictdiffer_Conflict(
            ('add', ['a', 'c', 0], [('d', '3')]),
            ('add', ['a', 'c', 0], [('d', '4')])
        )
    ]

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a.c.b': DictMergerOps.FALLBACK_KEEP_HEAD
    }

    expected_strategies = ['f', 's']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_custom_strategy_mixed_patches_with_nesting_after_list():
    patches = [
        Dictdiffer_Conflict(
            ('add', ['a', 'c', 0], [('d', '3')]),
            ('add', ['a', 'c', 0], [('d', '4')])
        ),
        Dictdiffer_Conflict(
            ('add', ['a', 'c', 0, 'b'], [('z', '5')]),
            ('add', ['a', 'c', 0, 'b'], [('z', '6')])
        )
    ]

    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE,
        'a.c.b': DictMergerOps.FALLBACK_KEEP_HEAD
    }

    expected_strategies = ['s', 'f']

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    strategies = [m._get_custom_strategy(patch) for patch in patches]

    assert expected_strategies == strategies


def test_get_rule_for_field_perfect_match():
    custom_ops = {
        'a.b.c.d': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    expected = 's'

    output = m._get_rule_for_field(['a', 'b', 'c', 'd'])

    assert expected == output


def test_get_rule_for_field_hierarchy_match():
    custom_ops = {
        'a.b': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    expected = 's'

    output = m._get_rule_for_field(['a', 'b', 'c', 'd'])

    assert expected == output


def test_get_rule_for_field_no_match():
    custom_ops = {
        'a.b': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops
    )

    expected = 'f'

    output = m._get_rule_for_field(['a', 'l', 'c', 'd'])

    assert expected == output


def test_get_rule_for_field_uses_key_path():
    custom_ops = {
        'a.b.c': DictMergerOps.FALLBACK_KEEP_UPDATE,
    }

    m = SkipListsMerger(
        {}, {}, {},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops, key_path=['a', 0]
    )

    expected = 's'

    output = m._get_rule_for_field(['b', 'c'])

    assert expected == output


def test_get_rule_for_field_looks_at_correct_path():
    def keep_spam(head, update, down_path):
        return 's' if update == 'spam' else 'f'

    custom_ops = {
        'a.b': keep_spam
    }
    head = {
        'a': {
            'b': [
                'egg',
                'bacon',
            ],
            'c': 'egg',
        },
    }
    update = {
        'a': {
            'b': [
                'spam',
                'bread',
            ],
        },
    }

    m = SkipListsMerger(
        {}, head, update,
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops, data_lists='a.b'
    )

    expected = 's'

    output = m._get_rule_for_field(['a', 'b', 0])

    assert expected == output


def test_merge_uses_custom_rules_for_dicts():
    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    m = SkipListsMerger(
        {}, {'a': 'head'}, {'a': 'update'},
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops,
    )
    try:
        m.merge()
    except MergeError:
        pass

    expected = {'a': 'update'}
    result = m.merged_root

    assert expected == result


def test_merge_uses_custom_rules_for_base_values():
    custom_ops = {
        'a': DictMergerOps.FALLBACK_KEEP_UPDATE
    }

    m = SkipListsMerger(
        '', 'head', 'update',
        DictMergerOps.FALLBACK_KEEP_HEAD,
        custom_ops=custom_ops, key_path=['a'],
    )
    try:
        m.merge()
    except MergeError:
        pass

    expected = 'update'
    result = m.merged_root

    assert expected == result


def test_merge_handles_duplicated_remove_patches():
    root = {'foo': 'bar', 'baz': 'spam'}
    head = {'foo': 'bar'}
    update = {'foo': 'bar'}

    m = SkipListsMerger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD)
    m.merge()

    assert m.merged_root == {'foo': 'bar'}



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Module for merging JSON objects.

To use this module you need to first import the main class:

>>> from json_merger import Merger

Then, import the configuration options:

>>> from json_merger.config import UnifierOps, DictMergerOps

The Basic Use Case
------------------

Let's assume we have JSON records that don't have any list fields --
They have string keys and as values other objects or primitive types.
In order to perform a merge we assume we have a lowest common ancestor
(``root``), a current version (``head``) and another version wich we want to
integrate into our record (``update``).

>>> root = {'name': 'John'} # A common ancestor of our person record
>>> head = {'name': 'Johnny', 'age': 32} # The current version of the record.
>>> update = {'name': 'Jonathan', 'address': 'Home'} # An updated version.

In this case we want to use the merger to compute one of the possible versions.

We create a merger instance in which we provide the default operation for
non-list fields and the one for list fields.

>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)
...            # Ignore UnifierOps for now.
>>> # We might get some exceptions
>>> from json_merger.errors import MergeError
>>> try:
...     m.merge()
... except MergeError:
...     pass # We don't care about this now.
>>> m.merged_root == {
...     'name': 'Johnny',
...     'age': 32,
...     'address': 'Home',
... }
True

The merged version kept the ``age`` field from the ``head`` object and the
``address`` field from the ``update`` object. The ``name`` field was different,
but because the strategy was ``FALLBACK_KEEP_HEAD`` the end result kept the
value from the ``head`` variable. To keep the ``update`` one, one can
use ``FALLBACK_KEEP_UPDATE``:

>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_UPDATE,
...            UnifierOps.KEEP_ONLY_HEAD_ENTITIES)
>>> rasised_something = False
>>> try:
...     m.merge()
... except MergeError:
...     raised_something = True
>>> m.merged_root == {
...     'name': 'Jonathan',
...     'age': 32,
...     'address': 'Home',
... }
True

If this type of conflict occurs, the merger will also populate a ``conflicts``
field. In this case the conflict holds the alternative name for our record.
Also, because a conflict occurred, the merge method also raised a MergeError.

For all the types of conflict that can be raised by the ``merge`` method
also check the :class:`json_merger.conflict.ConflictType` documentation.

>>> from json_merger.conflict import Conflict, ConflictType
>>> m.conflicts[0] == Conflict(ConflictType.SET_FIELD, ('name', ), 'Johnny')
True
>>> raised_something
True

Merging Lists With Base Values
------------------------------

For this example we are going to assume we want to merge sets of badges
that a person can receive.

>>> root = {'badges': ['bad', 'random']}
>>> head = {'badges': ['cool', 'nice', 'random']}
>>> update = {'badges': ['fun', 'nice', 'healthy']}

The most simple options are to either keep only the badges available in head
or only the badges available in the update. This can be done by specifying one
of:

  * ``UnifierOps.KEEP_ONLY_HEAD_ENTITIES``
  * ``UnifierOps.KEEP_ONLY_UPDATE_ENTITIES``

>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_ONLY_HEAD_ENTITIES)
>>> m.merge() # No conflict here
>>> m.merged_root['badges'] == ['cool', 'nice', 'random']
True
>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
>>> m.merge()
>>> m.merged_root['badges'] == ['fun', 'nice', 'healthy']
True

If we want to do a union of the elements we can use:

  * ``UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST``
  * ``UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST``

>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)
>>> m.merge() # No conflict here
>>> m.merged_root['badges'] == ['cool', 'fun', 'nice', 'random', 'healthy']
True
>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST)
>>> m.merge()
>>> m.merged_root['badges'] == ['fun', 'cool', 'nice', 'healthy', 'random']
True

These options keep the order relations between the entities. For example,
both ``'fun'`` and ``'cool'`` were placed before the ``'nice'`` entity but
between them there isn't any restriction. In such cases, for
``KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST`` we first pick the elements
that occur only in the `head` list and for
``KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST`` we first pick the ones that
occur only in the `update` list. If no such ordering is possible we first
add the elements found in the prioritized list and then the remaining ones.
Also, the method will raise a REORDER conflict.

>>> m = Merger([], [1, 2, 5, 3], [3, 1, 2, 4],
...            DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)
>>> try:
...     m.merge()
... except MergeError:
...     pass
>>> m.merged_root == [1, 2, 5, 3, 4]
True
>>> m.conflicts == [Conflict(ConflictType.REORDER, (), None)]
True
>>> m = Merger([], [1, 2, 5, 3], [3, 1, 2, 4],
...            DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST)
>>> try:
...     m.merge()
... except MergeError:
...     pass
>>> m.merged_root == [3, 1, 2, 4, 5]
True
>>> m.conflicts == [Conflict(ConflictType.REORDER, (), None)]
True

In the case in which ``root`` is represented by the latest automatic update
of a record (e.g. crawling some metadata source),
``head`` by manual edits of ``root`` and ``update`` by a new automatic
update, we might want to preserve only the entities in ``update`` and
notify the user in case some entry from ``head`` was removed.

  * ``UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE``

>>> root = {'badges': ['bad', 'random']}
>>> head = {'badges': ['cool', 'nice', 'random']}
>>> update = {'badges': ['fun', 'nice', 'random', 'healthy']}
>>> m = Merger(root, head, update, DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE)
>>> try:
...     m.merge()
... except MergeError:
...     pass
>>> m.merged_root['badges'] == ['fun', 'nice', 'random', 'healthy']
True
>>> m.conflicts == [Conflict(ConflictType.ADD_BACK_TO_HEAD,
...                          ('badges', ), 'cool')]
True

In this case, only ``'cool'`` was added "manually" and removed by the update.


Merging Lists Of Objects
------------------------

Assume the most complex case in which we need to merge lists of objects which
can also contain nested lists.

>>> root = {
...     'people': [
...         {'name': 'John', 'age': 13},
...         {'name': 'Peter'},
...         {'name': 'Max'}
...     ]}
>>> head = {
...     'people': [
...         {'name': 'John', 'age': 14,
...          'group': {'id': 'grp01'},
...          'person_id': '42',
...          'badges': [{'t': 'b0', 'e': True}, {'t': 'b1'}, {'t': 'b2'}]},
...         {'name': 'Peter', 'age': 15,
...          'badges': [{'t': 'b0'}, {'t': 'b1'}, {'t': 'b2'}]},
...         {'name': 'Max', 'age': 16}
...     ]}
>>> update = {
...     'people': [
...         {'name': 'Max', 'address': 'work'},
...         {'name': 'Johnnie', 'address': 'home',
...          'group': {'id': 'GRP01'},
...          'person_id': '42',
...          'age': 15,
...          'badges': [{'t': 'b1'}, {'t': 'b2'}, {'t': 'b0', 'e': False}]},
...     ]}

First of all we would like to define how to person records represent the same
entity. In this demo data model we can say that two records represent the
same person if any of the following is true:

  * They have the same ``name``
  * They have the same lowercased group id AND the same person_id

Then we define two badges as equal if they have the same ``t`` attribute.

In order to define a custom mode of linking records you can add comparator
classes for any of the list fields via the coparators keyword argument.
To define a simple comparsion that checks field equality you
can use :class:`json_merger.comparator.PrimaryKeyComparator`

In this case the fields from above look like this:

>>> from json_merger.comparator import PrimaryKeyComparator
>>> class PersonComparator(PrimaryKeyComparator):
...     primary_key_fields = ['name', ['group.id', 'person_id']]
...     normalization_functions = {'group.id': str.lower}
>>> class BadgesComparator(PrimaryKeyComparator):
...     primary_key_fields = ['t']

Note:
    You need to use a comparator class and not a comparator instance when
    defining the equality of two objects.

Next we would like to define how to do the merging:

  * In case of conflict keep ``head`` values.
  * For every list try to keep only the update entities.
  * For the badges list keep both entities with priority to the ``update``
    values.

>>> comparators = {'people': PersonComparator,
...                'people.badges': BadgesComparator}
>>> list_merge_ops = {
...     'people.badges': UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST
... }
>>> m = Merger(root, head, update,
...            DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_ONLY_UPDATE_ENTITIES,
...            comparators=comparators,
...            list_merge_ops=list_merge_ops)
>>> try:
...     m.merge()
... except MergeError:
...     pass
>>> m.merged_root == {
...     'people': [
...         {'name': 'Max', 'address': 'work', 'age': 16},
...         {'name': 'Johnnie', # Only update edited it.
...          'address': 'home',
...          'group': {'id': 'grp01'}, # From KEEP_HEAD
...          'person_id': '42',
...          'age': 14, # From KEEP_HEAD
...          'badges': [{'t': 'b1'}, {'t': 'b2'},
...                     {'t': 'b0', 'e': True}], # From KEEP_HEAD
...         },
...     ]}
True

Merging Data Lists
------------------

If you want to merge arrays of raw data (that do not encode any entities),
you can use the ``data_lists`` keyword argument. This argument treats
list indices as dictionary keys.

>>> root = {'f': {'matrix': [[0, 0], [0, 0]]}}
>>> head = {'f': {'matrix': [[1, 1], [0, 0]]}}
>>> update = {'f': {'matrix': [[0, 0], [1, 1]]}}
>>> m = Merger(root, head, update,
...            DictMergerOps.FALLBACK_KEEP_HEAD,
...            UnifierOps.KEEP_ONLY_UPDATE_ENTITIES,
...            data_lists=['f.matrix'])
>>> m.merge()
>>> m.merged_root == {'f': {'matrix': [[1, 1], [1, 1]]}}
True

Extending Comparators
---------------------

The internal API uses classes that extend
:class:`json_merger.comparator.BaseComparator` in order to check the semantic
equality of JSON objects. The interals call the ``get_matches`` method which
is implemented in terms of the ``equals`` method.  The most simple method to
extend this class is to override the ``equals`` method.

>>> from json_merger.comparator import BaseComparator
>>> class CustomComparator(BaseComparator):
...     def equal(self, obj1, obj2):
...         return abs(obj1 - obj2) < 0.2
>>> comp = CustomComparator([1, 2], [1, 2, 1.1])
>>> comp.get_matches('l1', 0) # elements matching l1[0] from l2
[(0, 1), (2, 1.1)]

If you want to implement another type of asignment you an compute all the
mathes and store them in the ``matches`` set by overriding the
``process_lists`` method. You need to put pairs of matching indices between
l1 and l2.

>>> from json_merger.comparator import BaseComparator
>>> class CustomComparator(BaseComparator):
...     def process_lists(self):
...         self.matches.add((0, 0))
...         self.matches.add((0, 1))
>>> comp = CustomComparator(['foo', 'bar'], ['bar', 'foo'])
>>> comp.get_matches('l1', 0) # elements matching l1[0] from l2
[(0, 'bar'), (1, 'foo')]

[contrib] Distance Function Matching
------------------------------------

To implement fuzzy matching we also allow matching by using a distane
function. This ensures a 1:1 mapping betwen the entities by minimizing
the total distance between all linked entities. To mark two of them
as equal you can provide a threshold for that distance. (This is why
it's best to normalize it between 0 and 1). Also, for speeding
up the matching you also can hint possible matches by bucketing matching
elements using a normalization function. In the next example we would
match some points in the coordinate system, each of them lying in a specific
square. The distance that we are going to use is the euclidean distance.
We will normalize the points to their integer counterpart.

>>> from json_merger.contrib.inspirehep.comparators import (
...     DistanceFunctionComparator)
>>> from math import sqrt
>>> class PointComparator(DistanceFunctionComparator):
...     distance_function = lambda p1, p2: sqrt((p1[0] - p2[0]) ** 2 +
...                                             (p1[1] - p2[1]) ** 2)
...     normalization_functions = [lambda p: (int(p[0]), int(p[1]))]
...     threshold = 0.5
>>> l1 = [(1.1, 1.1), (1.2, 1.2), (2.1, 2.1)]
>>> l2 = [(1.11, 1.11), (1.25, 1.25), (2.15, 2.15)]
>>> comp = PointComparator(l1, l2)
>>> comp.get_matches('l1', 0) # elements matching l1[0] from l2
[(0, (1.11, 1.11))]
>>> # match only the closest element, not everything under threshold.
>>> comp.get_matches('l1', 1)
[(1, (1.25, 1.25))]
>>> comp.get_matches('l1', 2)
[(2, (2.15, 2.15))]

[contrib] Custom Person Name Distance
-------------------------------------

We also provide a person name distance based on edit distance normalized
between 0 and 1. You just need to provide a function for tokenizing a full
name into NameToken or NameInitial - check ``simple_tokenize`` in the
contrib directory. This distance function matches initials with full
regular tokens and works with any name permutation. Also, this distance
calculator assumes the full name is inside the ``full_name`` field of a
dictionary. If you have the name in a different field you can just override
the class and call ``super`` on objects having the name in the ``full_name``
field.

>>> from json_merger.contrib.inspirehep.author_util import (
...     AuthorNameDistanceCalculator, simple_tokenize)
>>> dst = AuthorNameDistanceCalculator(tokenize_function=simple_tokenize)
>>> dst({'full_name': u'Doe, J.'}, {'full_name': u'John, Doe'}) < 0.1
True

Also we have functions for normalizing an author name with different
heuristics to speed up the distance function matching.

>>> from json_merger.contrib.inspirehep.author_util import (
...     AuthorNameNormalizer)
>>> identity = AuthorNameNormalizer(simple_tokenize)
>>> identity({'full_name': 'Doe, Johnny Peter'})  # doctest: +SKIP
('doe', 'johnny', 'peter')
>>> asciified = AuthorNameNormalizer(simple_tokenize,
...                                  asciify=True)
>>> asciified({'full_name': 'Dœ, Jöhnny Péter'})  # doctest: +SKIP
('doe', 'johnny', 'peter')
>>> one_fst_name = AuthorNameNormalizer(simple_tokenize,
...                                     first_names_number=1)
>>> one_fst_name({'full_name': 'Doe, Johnny Peter'})  # doctest: +SKIP
('doe', 'johnny')
>>> last_name_one_initial = AuthorNameNormalizer(simple_tokenize,
...                                              first_names_number=1,
...                                              first_name_to_initial=True)
... # doctest: +SKIP
>>> last_name_one_initial({'full_name': 'Doe, Johnny Peter'})  # doctest: +SKIP
('doe', 'j')

These instances can be used as class parameters for
``DistanceFunctionComparator``
"""

from __future__ import absolute_import, print_function

from .merger import Merger
from .version import __version__

__all__ = ('__version__', 'Merger')



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Test utils."""

from __future__ import absolute_import, print_function

import pytest

from json_merger.utils import (
    del_obj_at_key_path, get_obj_at_key_path, set_obj_at_key_path,
    get_conf_set_for_key_path, remove_prefix, force_list)


def test_del_obj_at_key_path():
    o = {'a': [{'a': [1, 2, 3]}]}
    del_obj_at_key_path(o, ['a', 0, 'a', 1])
    assert o == {'a': [{'a': [1, 3]}]}

    del_obj_at_key_path(o, ['a', 0, 5], False)
    assert o == {'a': [{'a': [1, 3]}]}

    with pytest.raises(KeyError):
        del_obj_at_key_path(o, ['a', 0, 5])
    with pytest.raises(KeyError):
        del_obj_at_key_path(o, ['a', 0, 'a', 0, 1])

    # Now destroy the object
    del_obj_at_key_path(o, ['a', 0, 'a'])
    assert o == {'a': [{}]}
    del_obj_at_key_path(o, ['a', 0])
    assert o == {'a': []}
    del_obj_at_key_path(o, ['a'])
    assert o == {}


def test_get_obj_at_key_path():
    o = {'a': [{'a': [1, 2, 3]}]}
    o1 = get_obj_at_key_path(o, ['a', 0, 'a', 0])
    assert o1 == 1

    o2 = get_obj_at_key_path(o, ['a', 0, 'a'])
    assert o2 == [1, 2, 3]

    o3 = get_obj_at_key_path(o, ['a', 0])
    assert o3 == {'a': [1, 2, 3]}

    o4 = get_obj_at_key_path(o, ['a'])
    assert o4 == [{'a': [1, 2, 3]}]

    o5 = get_obj_at_key_path(o, [])
    assert o5 == o

    o6 = get_obj_at_key_path(o, ['a', '123'])
    assert o6 is None


def test_set_obj_at_key_path():
    o = {'a': [{'a': [1, 2, 3]}]}
    original_o = {'a': [{'a': [1, 2, 3]}]}

    with pytest.raises(KeyError):
        set_obj_at_key_path(o, ['a', '1234'], 42)
    o = set_obj_at_key_path(o, ['a', '1234'], 42, False)
    assert o == original_o

    with pytest.raises(KeyError):
        set_obj_at_key_path(o, ['a', 1234], 42)
    o = set_obj_at_key_path(o, ['a', 1234], 42, False)
    assert o == original_o

    with pytest.raises(KeyError):
        set_obj_at_key_path(o, ['a', 0, 'd', 10, 11, 12], 42)
    o = set_obj_at_key_path(o, ['a', 0, 'd', 10, 11, 12], 42, False)
    assert o == original_o

    o = set_obj_at_key_path(o, ['a', 0, 'a', 0], 42)
    assert o['a'][0]['a'][0] == 42
    o = set_obj_at_key_path(o, ['a', 0, 'a'], 42)
    assert o['a'][0]['a'] == 42
    o = set_obj_at_key_path(o, ['a', 0], 42)
    assert o['a'][0] == 42
    o = set_obj_at_key_path(o, ['a'], 42)
    assert o['a'] == 42
    o = set_obj_at_key_path(o, [], 42)
    assert o == 42


def test_get_conf_set_for_key_path():
    expected = set(['d', 'd.e', 'd.e.f'])
    actual = get_conf_set_for_key_path(
        set(['a.b.c.d', 'a.b.c.d.e', 'a.b.c.d.e.f',
             'd', 'd.e', 'd.e.f']), ('a', 'b', 'c'))
    assert actual == expected


def test_remove_prefix_value_error():
    with pytest.raises(ValueError):
        remove_prefix('a.b.c.d.e', 'a.b.c.e')


@pytest.mark.parametrize('value, expected', [
    (None, [None]),
    ('foo', ['foo']),
    (('foo', 'bar'), ['foo', 'bar']),
    (['foo', 'bar'], ['foo', 'bar'])
])
def test_force_list(value, expected):
    result = force_list(value)
    assert result == expected



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Acceptance scenarios for the merger."""

from __future__ import absolute_import, print_function

import pytest

from json_merger import Merger
from json_merger.config import DictMergerOps, UnifierOps
from json_merger.errors import MergeError
from json_merger.contrib.inspirehep.comparators import (
        DistanceFunctionComparator)
from json_merger.contrib.inspirehep.author_util import (
        simple_tokenize, AuthorNameDistanceCalculator, AuthorNameNormalizer)
from json_merger.comparator import PrimaryKeyComparator

from json_merger.conflict import Conflict, ConflictType


class TitleComparator(PrimaryKeyComparator):

    primary_key_fields = ['source']


class AffiliationComparator(PrimaryKeyComparator):

    primary_key_fields = ['value']


class AuthorComparator(DistanceFunctionComparator):
    norm_functions = [
        # Better hints can be given by normalizing by primary key,
        # (e.g. recid, orcid, ...)
        # but this type of normalization is not implemented in contrib.
        AuthorNameNormalizer(simple_tokenize),
        AuthorNameNormalizer(simple_tokenize, 1),
        AuthorNameNormalizer(simple_tokenize, 1, True)
    ]
    distance_function = AuthorNameDistanceCalculator(simple_tokenize)
    threshold = 0.12


COMPARATORS = {
    'authors': AuthorComparator,
    'authors.affiliations': AffiliationComparator,
    'titles': TitleComparator
}

LIST_MERGE_OPS = {
    'titles': UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST,
    'authors.affiliations': UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST
}


def _deserialize_conflict(conflict_type, path, body):
    if conflict_type == ConflictType.MANUAL_MERGE:
        body = tuple(body)
    return Conflict(conflict_type, tuple(path), body)


@pytest.mark.parametrize('scenario', [
    'author_typo_update_fix',
    'author_typo_curator_fix',
    'author_typo_update_and_curator_fix',
    'author_typo_conflict',
    'author_prepend_and_curator_typo_fix',
    'author_delete_and_single_curator_typo_fix',
    'author_delete_and_double_curator_typo_fix',
    'author_reorder_and_double_curator_typo_fix',
    'author_reorder_conflict',
    'author_replace_and_single_curator_typo_fix',
    'author_delete_and_double_curator_typo_fix',
    'author_curator_collab_addition',
    'author_affiliation_addition',
    'author_double_match_conflict',
    'author_double_match_unambiguous_fix',
    'title_addition',
    'title_change'
])
def test_author_typo_scenarios(update_fixture_loader, scenario):
    root, head, update, exp, desc = update_fixture_loader.load_test(scenario)
    merger = Merger(root, head, update,
                    DictMergerOps.FALLBACK_KEEP_HEAD,
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES,
                    comparators=COMPARATORS,
                    list_merge_ops=LIST_MERGE_OPS)
    if exp.get('conflicts'):
        with pytest.raises(MergeError) as excinfo:
            merger.merge()
        expected_conflicts = [_deserialize_conflict(t, p, b)
                              for t, p, b in exp.pop('conflicts')]
        assert set(expected_conflicts) == set(excinfo.value.content)
    else:
        merger.merge()

    assert merger.merged_root == exp, desc


def test_add_author_in_head(update_fixture_loader):

    root, head, update, exp, desc = update_fixture_loader.load_test(
        'author_added_only_in_head'
    )
    list_config = {
        'authors':
            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE
    }

    merger = Merger(root, head, update,
                    DictMergerOps.FALLBACK_KEEP_HEAD,
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES,
                    comparators=COMPARATORS,
                    list_merge_ops=list_config)
    merger.merge()
    assert merger.merged_root == exp, desc


def test_author_deleted_in_update(update_fixture_loader):

    root, head, update, exp, desc = update_fixture_loader.load_test(
        'author_added_only_in_head'
    )
    list_config = {
        'authors':
            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE
    }

    merger = Merger(root, head, update,
                    DictMergerOps.FALLBACK_KEEP_HEAD,
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES,
                    comparators=COMPARATORS,
                    list_merge_ops=list_config)
    merger.merge()
    assert merger.merged_root == exp, desc



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Python module that is able to merge json record objects."""

import os
import sys

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.25',
    'coverage>=4.0',
    'isort>=4.2.2',
    'pydocstyle>=1.0.0',
    'pytest-cache>=1.0',
    'pytest-cov>=1.8.0',
    'pytest>=4.6.11',
    'flake8>=3.9.0',
]

contrib_require = [
    'editdistance>=0.3.1',
    'munkres<=1.0.12',
    'Unidecode==0.4.19' if sys.version_info < (3, 6) else 'Unidecode>=0.4.19'
]

tests_require += contrib_require

extras_require = {
    'contrib': contrib_require,
    'docs': [
        'Sphinx>=1.4.2',
    ],
    'tests': tests_require
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

install_requires = [
    'dictdiffer==0.8.1' if sys.version_info < (3, 6) else 'dictdiffer>=0.6.0',
    'six>=1.10.0',
    "autosemver==0.5.5",
    'pyrsistent>=0.11.13'
]

packages = find_packages()

setup_require = [
    "autosemver==0.5.5"
]

setup(
    name='json-merger',
    autosemver=True,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='JSON patch merge conflict',
    license='GPLv2',
    author='CERN',
    author_email='admin@inspirehep.net',
    url='https://github.com/inveniosoftware-contrib/json-merger',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
    },
    setup_requires=setup_require,
    extras_require=extras_require,
    install_requires=install_requires,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 4 - Beta',
    ],
)
